class Search {
    /** @type {HTMLDivElement} */
    #searchResultParent = document.getElementById('search-result-parent');
    /** @type {HTMLDivElement} */
    #searchResultTracks = document.getElementById('search-result-tracks');
    /** @type {HTMLDivElement} */
    #searchResultArtists = document.getElementById('search-result-artists');
    /** @type {HTMLDivElement} */
    #searchResultAlbums = document.getElementById('search-result-albums');
    /** @type {HTMLInputElement} */
    #queryInput =  document.getElementById('search-query');
    /** @type {number} */
    #searchTimeoutId = null;

    constructor() {
        eventBus.subscribe(MusicEvent.METADATA_CHANGE, () => {
            if (!windows.isOpen('window-search')) {
                console.debug('search: ignore METADATA_CHANGE, search window is not open');
                return;
            }
            console.debug('search: search again after receiving METADATA_CHANGE event');
            this.#performSearch(true);
        });

        document.addEventListener('DOMContentLoaded', () => {
            this.#queryInput.addEventListener('input', () => this.#performSearch());

            document.getElementById('open-window-search').addEventListener('click', () => {
                /** @type {HTMLInputElement} */
                const queryField = document.getElementById('search-query');
                queryField.value = '';
                setTimeout(() => queryField.focus({focusVisible: true}), 50); // high delay is necessary, I don't know why
                this.#searchResultParent.hidden = true;
            });
        });
    }

    async #performSearch(searchNow = false) {
        // Only start searching after user has finished typing for better performance and fewer network requests
        if (!searchNow) {
            // Reset timer when new change is received
            if (this.#searchTimeoutId != null) {
                clearTimeout(this.#searchTimeoutId);
            }
            // Perform actual search in 200 ms
            this.#searchTimeoutId = setTimeout(() => this.#performSearch(true), 200);
            return;
        }

        const query = this.#queryInput.value;

        /** @type {Array<Track>} */
        let tracks = [];
        /** @type {Array<Album>} */
        let albums = [];
        if (query.length > 1) {
            const result = await music.search(query);
            tracks = result.tracks;
            albums = result.albums;
        } else {
            tracks = [];
            albums = [];
        }

        if (tracks.length == 0) {
            this.#searchResultParent.hidden = true;
            return;
        }

        this.#searchResultParent.hidden = false;

        // Tracks
        {
            this.#searchResultTracks.replaceChildren();
            await browse.appendTrackRows(this.#searchResultTracks, tracks);
        }

        // Artists
        {
            const table = document.createElement('table');
            const listedArtists = new Set(); // to prevent duplicates, but is not actually used to preserve ordering
            for (const track of tracks) {
                if (track.artists.length == 0) {
                    continue;
                }

                for (const artist of track.artists) {
                    if (listedArtists.has(artist)) {
                        continue;
                    }
                    listedArtists.add(artist);

                    const artistLink = document.createElement('a');
                    artistLink.textContent = artist;
                    artistLink.onclick = () => browse.browseArtist(artist);

                    const td = document.createElement('td');
                    td.append(artistLink);
                    const row = document.createElement('tr');
                    row.append(td);
                    table.append(row);
                }
            }

            this.#searchResultArtists.replaceChildren(table);
        }

        // Albums
        {
            const newChildren = [];
            for (const album of albums) {
                const img = document.createElement('div');
                img.classList.add('search-result-album');
                const imgUri = await album.getCover('low', true);
                img.style.background = `black url("${imgUri}") no-repeat center`;
                img.style.backgroundSize = 'cover';
                img.onclick = () => browse.browseAlbum(album.name, album.artist);
                newChildren.push(img);
                if (newChildren.length > 6) {
                    break;
                }
            }

            this.#searchResultAlbums.replaceChildren(...newChildren);
        }
    }
}

const search = new Search();
